
class Column:
    dtype = None
    sql_string = None

    def __init__(self, dtype=None, sql_string=None):
        self.dtype = dtype
        self.sql_string = sql_string
    
    def __copy__(self):
        return type(self)(self.sql_string)



import pandas as pd
from pandas_to_sql.utils.pandas_dataframe_intercepter import PandasDataFrameIntercepter
from copy import copy

def flatten_grouped_dataframe(df):
    if not isinstance(df, PandasDataFrameIntercepter):
        raise Exception(f"can only get type {str(type(PandasDataFrameIntercepter))}")
    
    df_c = copy(df.df_pandas)
    if isinstance(df_c, pd.core.series.Series):
        series_name = df_c.name
        new_col_name = list(filter(lambda k: k.startswith(series_name), df.df_sql_convert_table.columns.keys()))[0]
        df_c = df_c.reset_index().rename(columns={series_name: new_col_name})
    else:
        df_c.columns = df_c.columns.map('_'.join)
        df_c = df_c.reset_index()
    return PandasDataFrameIntercepter(df_c, copy(df.df_sql_convert_table))



from datetime import timedelta
from dateutil.relativedelta import relativedelta
import pandas as pd
from pandas_to_sql.engine.columns.column import Column
from pandas_to_sql.engine.columns.common import add_comparison_operators_to_class, value_to_sql_string
from pandas_to_sql.engine.columns.numeric_columns import IntColumn, FloatColumn

time_unit_to_format = {
    'second': '%S',
    'month': '%m',
    'minute': '%M',
    'hour': '%H',
    'week': '%W',
    'year': '%Y',
    'day': '%d',
    'dayofweek': '%w'}


class TimestampColumn(Column):

    def __init__(self, sql_string):
        super().__init__(dtype='TIMESTAMP', sql_string=sql_string)

    def __getattribute__(self, attr):
        if attr == 'dt':
            return self
        if attr == 'dayofweek':
            return self.extract_dayofweek()
        if attr in time_unit_to_format.keys():
            return self.extract_time_unit(time_unit_to_format[attr])

        return object.__getattribute__(self, attr)


    def get_sql_for_format(self, format):
        return f"(CAST(strftime('{format}', {value_to_sql_string(self)}) AS INT))"
    
    def extract_time_unit(self, format):
        sql_string = self.get_sql_for_format(format)
        return IntColumn(sql_string=sql_string)
    
    def extract_dayofweek(self):
        sql_string = self.get_sql_for_format(time_unit_to_format['dayofweek'])
        sql_string = f"( ({sql_string} + 6) % 7)"
        return IntColumn(sql_string=sql_string)


def __my_add__(col, v):
    if isinstance(v, timedelta):
        # https://docs.python.org/3/library/datetime.html#datetime.timedelta
        sign = '+' if v.days >= 0 else '-'
        added_days = f"'{sign}{abs(v.days)} days'"

        sign = '+' if v.seconds >= 0 else '-'
        added_seconds = f"'{sign}{abs(v.seconds)} seconds'"

        sql_string = f"(datetime({value_to_sql_string(col)}, {added_days}, {added_seconds}))"
        return TimestampColumn(sql_string=sql_string)
    elif isinstance(v, relativedelta):
        s = []
        for t_type, t_value in v.kwds.items():
            sign = '+' if t_value >= 0 else '-'
            s.append(f"'{sign}{abs(t_value)} {t_type}'")
        sql_string = f"(datetime({value_to_sql_string(col)}, {', '.join(s)}))"
        return TimestampColumn(sql_string=sql_string)
    elif isinstance(v, pd.offsets.DateOffset):
        s = []
        for t_type, t_value in v.kwds.items():
            sign = '+' if t_value >= 0 else '-'
            s.append(f"'{sign}{abs(t_value)} {t_type}'")
        sql_string = f"(datetime({value_to_sql_string(col)}, {', '.join(s)}))"
        return TimestampColumn(sql_string=sql_string)
    else:
        raise Exception(f'Supporting only timedelta, got {str(type(v))}')


add_comparison_operators_to_class(TimestampColumn)

TimestampColumn.__add__ = __my_add__
TimestampColumn.__radd__ = lambda self, l: __my_add__(self, l)
TimestampColumn.__sub__ = lambda self, r: __my_add__(self, -r)




import pandas as pd
from pandas.testing import assert_frame_equal
import pytest


def assert_dataframes_equals(expected, actual):
    assert expected.shape==actual.shape
    assert set(expected.columns) == set(actual.columns)
    columns_order = list(expected.columns)
    a = actual[columns_order].sort_values(by=list(actual.columns)).reset_index(drop=True)
    e = expected[columns_order].sort_values(by=list(actual.columns)).reset_index(drop=True)
    assert_frame_equal(e, a, check_dtype=False)


def get_expected_and_actual(df):
    actual_query_string = df.df_sql_convert_table.get_sql_string()
    actual_columns = df.df_sql_convert_table.columns
    timestamp_columns = [c for c in actual_columns.keys() if actual_columns[c].dtype == 'TIMESTAMP']

    df_actual = pd.read_sql_query(actual_query_string, pytest.sql_connection, parse_dates=timestamp_columns)
    df_expected = df.df_pandas

    return df_expected, df_actual

def assert_(df):
    df_expected, df_actual = get_expected_and_actual(df)

    # i = df_expected.new_value != df_actual.new_value
    # a=df_expected[i][:3]
    # b=df_expected[i][:3]

    assert_dataframes_equals(df_expected, df_actual)



from pandas_to_sql.engine.columns.column import Column
from pandas_to_sql.engine.columns.common import value_to_sql_string, add_comparison_operators_to_class


class BoolColumn(Column):
    def __init__(self, sql_string):
        super().__init__(dtype='BOOL', sql_string=sql_string)
    
    def __neg__(self):
        return BoolColumn(sql_string=f'(NOT({value_to_sql_string(self)}))')

    def __invert__(self):
        return BoolColumn(sql_string=f'(NOT({value_to_sql_string(self)}))')


add_comparison_operators_to_class(BoolColumn)


import pandas as pd


## Types
def convert_df_type(col_type):
    if pd.api.types.is_bool_dtype(col_type): return 'BOOL'
    elif pd.api.types.is_integer_dtype(col_type): return 'INT'
    elif pd.api.types.is_numeric_dtype(col_type): return 'FLOAT'
    elif pd.api.types.is_string_dtype(col_type): return 'VARCHAR'
    elif pd.api.types.is_datetime64_any_dtype(col_type): return 'TIMESTAMP'
    else: raise Exception(f"could not convert column type. got: {str(col_type)}")


def create_schema_from_df(df):        
    schema = {}
    for col_name, col_type in df.dtypes.items():
        schema[col_name] = convert_df_type(col_type)
    return schema


from pandas_to_sql.engine.columns.column import Column
from pandas_to_sql.engine.columns.numeric_columns import IntColumn
from pandas_to_sql.engine.columns.common import add_comparison_operators_to_class, value_to_sql_string, create_column_from_operation


class StrColumn(Column):
    def __init__(self, sql_string):
        super().__init__(dtype='VARCHAR', sql_string=sql_string)

    def __getattribute__(self, attr):
        if attr == 'str':
            return self
        return object.__getattribute__(self, attr)
    
    def __add__(self, r):
        return create_column_from_operation(self, r, StrColumn, '||')

    def __radd__(self, l):
        return create_column_from_operation(self, StrColumn, l, '||')


add_comparison_operators_to_class(StrColumn)

StrColumn.lower = lambda self: StrColumn(sql_string=f'(LOWER({value_to_sql_string(self)}))')
StrColumn.upper = lambda self: StrColumn(sql_string=f'(UPPER({value_to_sql_string(self)}))')

StrColumn.replace = lambda self, old, new: \
    StrColumn(sql_string=f'(REPLACE({value_to_sql_string(self)}, {value_to_sql_string(old)}, {value_to_sql_string(new)}))')


def slice_(self, start=None, stop=None,j=None):
    if j: raise 'slice "step" not supported'

    start = start if start else 0
    start+=1

    if stop:
        stop +=1
        length = stop - start
        s = f'(SUBSTR({value_to_sql_string(self)}, {start}, {length}))'
    else:
        s = f'(SUBSTR({value_to_sql_string(self)}, {start}))'

    return StrColumn(sql_string=s)


StrColumn.slice = slice_
    



def strip_(self, op, chars=None):
    if not chars:
        chars = ' '
    if not isinstance(chars, str):
        raise f'"chars" must be str. got {str(type(chars))}'
    
    s = f"({op}({value_to_sql_string(self)}, {value_to_sql_string(chars)}))"
    return StrColumn(sql_string=s)


StrColumn.strip = lambda self, chars=None: strip_(self, 'TRIM', chars)
StrColumn.lstrip = lambda self, chars=None: strip_(self, 'LTRIM', chars)
StrColumn.rstrip = lambda self, chars=None: strip_(self, 'RTRIM', chars)

StrColumn.len = lambda self: IntColumn(sql_string=f'(LENGTH({value_to_sql_string(self)}))')



def contains(self, s, case=True):
    if not isinstance(s, str):
        raise f'"s" must be str. got {str(type(s))}'
    
    if case==False:
        sql_string = f"(INSTR(LOWER({value_to_sql_string(self)}), LOWER({value_to_sql_string(s)})))"
    else:
        sql_string = f"(INSTR({value_to_sql_string(self)}, {value_to_sql_string(s)}))"
    
    # sql_string = f"(INSTR({value_to_sql_string(self)}, {value_to_sql_string(s)}))"
    sql_string = f"(CAST({sql_string} > 0 AS BOOL))"
    return StrColumn(sql_string=sql_string)


StrColumn.contains = contains



from pandas_to_sql.engine.columns.column import Column
from pandas_to_sql.engine.columns.common import add_comparison_operators_to_class, value_to_sql_string, create_column_from_operation


class FloatColumn(Column):
    def __init__(self, sql_string):
        super().__init__(dtype='FLOAT', sql_string=sql_string)


class IntColumn(Column):
    def __init__(self, sql_string):
        super().__init__(dtype='INT', sql_string=sql_string)


def __floordiv__(self, r):
    # http://sqlite.1065341.n5.nabble.com/floor-help-td46158.html
    return FloatColumn(sql_string=f'( ROUND(({value_to_sql_string(self)} / {value_to_sql_string(r)}) - 0.5) )')

def __rfloordiv__(self, l):
    # http://sqlite.1065341.n5.nabble.com/floor-help-td46158.html
    return FloatColumn(sql_string=f'( ROUND(({value_to_sql_string(l)} / {value_to_sql_string(self)}) - 0.5) )')

def is_int(v):
    return isinstance(v, int) or isinstance(v, IntColumn)

def numeric_op_result_from_types(l, r):
    x = IntColumn if is_int(l) and is_int(r) else FloatColumn
    return x

def __add__(self, r):
    res_column_type = numeric_op_result_from_types(self, r)
    return create_column_from_operation(self, r, res_column_type, '+')

def __radd__(self, l):
    res_column_type = numeric_op_result_from_types(l, self)
    return create_column_from_operation(l, self, res_column_type, '+')

def __sub__(self, r):
    res_column_type = numeric_op_result_from_types(self, r)
    return create_column_from_operation(self, r, res_column_type, '-')

def __rsub__(self, l):
    res_column_type = numeric_op_result_from_types(l, self)
    return create_column_from_operation(l, self, res_column_type, '-')

def __mul__(self, r):
    res_column_type = numeric_op_result_from_types(self, r)
    return create_column_from_operation(self, r, res_column_type, '*')

def __rmul__(self, l):
    res_column_type = numeric_op_result_from_types(l, self)
    return create_column_from_operation(l, self, res_column_type, '*')

def __truediv__(self, r):
    return FloatColumn(sql_string=f'(({value_to_sql_string(self)} + 0.0) / {value_to_sql_string(r)})')  

def __rtruediv__(self, l):
    return FloatColumn(sql_string=f'(({value_to_sql_string(l)} + 0.0) / {value_to_sql_string(self)})')  

def __abs__(self):
    return type(self)(sql_string=f'ABS({value_to_sql_string(self)})')  

def __neg__(self):
    return type(self)(sql_string=f'(-({value_to_sql_string(self)}))')


def round_(self):
    # https://docs.python.org/3/library/functions.html#round
    v = value_to_sql_string(self)
    integer_part = f'(CAST({v} AS INT))'
    fractional_part = f'(ABS({v}) - ROUND(ABS({v})-0.5))'

    is_integer_part_even = f'({integer_part}%2 == 0)'
    is_fractional_part_exactly_half = f'({fractional_part}==.5)'

    simple_round = f'(ROUND({v}))'
    round_with_change = f'(CASE WHEN {v}>0 THEN ROUND({v}-0.001) ELSE ROUND({v}+0.001) END)'

    s = f'(CASE WHEN {is_fractional_part_exactly_half} AND {is_integer_part_even} THEN {round_with_change} ELSE {simple_round} END)'
    return FloatColumn(sql_string=s)



add_comparison_operators_to_class(FloatColumn)
FloatColumn.__add__ = __add__
FloatColumn.__radd__ = __radd__
FloatColumn.__sub__ = __sub__
FloatColumn.__rsub__ = __rsub__
FloatColumn.__mul__ = __mul__
FloatColumn.__rmul__ = __rmul__
FloatColumn.__floordiv__ = __floordiv__
FloatColumn.__rfloordiv__ = __rfloordiv__
FloatColumn.__truediv__ = __truediv__
FloatColumn.__rtruediv__ = __rtruediv__
FloatColumn.__abs__ = __abs__
FloatColumn.__neg__ = __neg__
FloatColumn.round = round_


add_comparison_operators_to_class(IntColumn)
IntColumn.__add__ = __add__
IntColumn.__radd__ = __radd__
IntColumn.__sub__ = __sub__
IntColumn.__rsub__ = __rsub__
IntColumn.__mul__ = __mul__
IntColumn.__rmul__ = __rmul__
IntColumn.__floordiv__ = __floordiv__
IntColumn.__rfloordiv__ = __rfloordiv__
IntColumn.__truediv__ = __truediv__
IntColumn.__rtruediv__ = __rtruediv__
IntColumn.__abs__ = __abs__
IntColumn.__neg__ = __neg__
IntColumn.round = round_




import numbers
import operator
from datetime import datetime
from pandas_to_sql.engine.columns.column import Column


def get_column_class_from_type(col_type):
    from pandas_to_sql.engine.columns.bool_column import BoolColumn
    from pandas_to_sql.engine.columns.numeric_columns import IntColumn, FloatColumn
    from pandas_to_sql.engine.columns.str_column import StrColumn
    from pandas_to_sql.engine.columns.timestamp_column import TimestampColumn
    if col_type == 'BOOL': return BoolColumn
    elif col_type == 'INT': return IntColumn
    elif col_type == 'FLOAT': return FloatColumn
    elif col_type == 'VARCHAR': return StrColumn
    elif col_type == 'TIMESTAMP': return TimestampColumn
    else: raise Exception(f"could not convert column type. got: {str(col_type)}")


def value_to_sql_string(value):
    if isinstance(value, numbers.Number):
        return str(value)
    elif isinstance(value, str):
        return "'" + value + "'"
    elif isinstance(value, datetime):
        return f"datetime('{value.strftime('%Y-%m-%d %H:%M:%S')}')"
    elif isinstance(value, Column):
        return value.sql_string
    raise Exception(f"Value not supported. supporting: premitives and {str(Column)}. got {str(type(value))}")


def create_column_from_value(v):
    from pandas_to_sql.engine.columns.bool_column import BoolColumn
    from pandas_to_sql.engine.columns.str_column import StrColumn
    from pandas_to_sql.engine.columns.timestamp_column import TimestampColumn
    from pandas_to_sql.engine.columns.numeric_columns import IntColumn, FloatColumn
    sql_string = value_to_sql_string(v)
    if isinstance(v, int): return IntColumn(sql_string)
    if isinstance(v, float): return FloatColumn(sql_string)
    if isinstance(v, str): return StrColumn(sql_string)
    if isinstance(v, bool): return BoolColumn(sql_string)
    if isinstance(v, datetime): return TimestampColumn(sql_string)
    
    raise Exception(f'trying to set table column with unsupported type. expected types are Column or primitives. got type: {str(type(newvalue))}' )

def create_column_from_operation(l, r, dtype, op):
    return dtype(sql_string=f'({value_to_sql_string(l)} {op} {value_to_sql_string(r)})')  


def add_comparison_operators_to_class(class_type):
    from pandas_to_sql.engine.columns.bool_column import BoolColumn

    def __lt__(self,other):
        return create_column_from_operation(self, other, BoolColumn, '<')

    def __le__(self,other):
        return create_column_from_operation(self, other, BoolColumn, '<=')

    def __gt__(self,other):
        return create_column_from_operation(self, other, BoolColumn, '>')

    def __ge__(self,other):
        return create_column_from_operation(self, other, BoolColumn, '>=')

    def __eq__(self,other):
        return create_column_from_operation(self, other, BoolColumn, '=')

    def __ne__(self,other):
        return create_column_from_operation(self, other, BoolColumn, '<>')
    
    def __and__(self,other):
        return create_column_from_operation(self, other, BoolColumn, 'AND')
    
    def __or__(self,other):
        return create_column_from_operation(self, other, BoolColumn, 'OR')

    class_type.__lt__ = __lt__
    class_type.__gt__ = __gt__
    class_type.__le__ = __le__
    class_type.__ge__ = __ge__
    class_type.__eq__ = __eq__
    class_type.__ne__ = __ne__
    class_type.__and__ = __and__
    class_type.__or__ = __or__
